#ifndef __disasm__
#define __disasm__

string instr_code_str(uint16_t instr_code)
{
    switch(instr_code)
    {
        case 0:return "ldl";// ldl
        case 1:return "ldh";// ldh
        case 2:return "mv2reg";// mv2reg
        case 3:return "mv2acc";// mv2acc
        case 4:return "rd";// rd
        case 5:return "wr";// wr
        case 6:return "rdInc";// rdInc
        case 7:return "wrInc";// wrInc
        case 8:return "rdDec";// rdDec
        case 9:return "wrDec";// wrDec
        case 10:return "jmp";// jmp
        case 11:return "jmr";// jmr, add Ri to PC
        case 12:return "jmi";// jmi
        case 13:return "jmri";// jmri
        case 14:return "jsr";// jsr
        case 15:return "jsri";// jsri
        case 16:return "rts";// rts
        case 17:return "nop";// nop
        case 18:return "add";// add
        case 19:return "sub";// sub
        case 20:return "mul";// mul
        case 21:return "cmp";//cmp
        case 22:return "swap";// swap
        case 23:return "clr";//clear
        case 24:return "and";//and
        case 25:return "or";//or
        case 26:return "xor";//xor
        case 27:return "not";//not
        case 28:return "lsr";//lsr
        case 29:return "lsl";//lsl
        case 30:return "ror";//ror
        case 31:return "rol";//rol
        default:return "<unk>";
    }
}

string disassemble(uint16_t instr)
{
    ostringstream oss;
    uint16_t instrCode  =  instr >> 11;
    uint16_t instrPred  = (instr >> 8) & 0x7;
    uint16_t instrFlag  = (instr >> 7) & 0x1;
    uint16_t reg = instr & 0x7f;
    uint32_t imm = instr & 0xff;
    
    if(instrFlag && instrCode>=18 && instrCode<=31)
        oss<<"!s ";
    
    switch(instrPred)
    {
        case 0: // GT
            oss<<"?p GT ";
            break;
        case 1: // GE
            oss<<"?p GE ";
            break;
        case 2: // EQ
            oss<<"?p EQ ";
            break;
        case 3: // NE
            oss<<"?p NE ";
            break;
        case 4: // LE
            oss<<"?p LE ";
            break;
        case 5: // LT
            oss<<"?p LT ";
            break;
        case 6: // OV
            oss<<"?p OV ";
            break;
        case 7: // NC
            break;
    }
    
    oss<<instr_code_str(instrCode);
    
    if(instrCode==0 || instrCode==1 || instrCode==12 || instrCode==13 || instrCode==15)
    {
        oss<<" 0x"<<hex<<imm;
    }
    else if(instrCode==17 || instrCode==16)
    {
    }
    else
    {
        oss<<" r"<<reg;
    }
    
    return oss.str();
}

#endif // __disasm__
