#include <cstdio>
#include <cstdlib>
#include <cassert>
#include <stdint.h>
#include <ctime>

int low(int a)
{
    return a&0xff;
}

int high(int a)
{
    return (a>>8)&0xff;
}

bool execute_and_test(int x,int y,const char *instr,const char *flag,bool expected_result)
{
    const char *asm_filename="__temp__.asm";
    const char *obj_filename="__temp__.obj";
    const char *out_filename="__temp__.out";
    FILE *f=fopen(asm_filename,"w");
    
    fprintf(f,"      ldl %#x\n",low(y));
    fprintf(f,"      ldh %#x\n",high(y));
    fprintf(f,"      mv2reg r1\n");
    fprintf(f,"      ldl %#x\n",low(x));
    fprintf(f,"      ldh %#x\n",high(x));
    fprintf(f,"   !s %s r1\n",instr);
    fprintf(f,"      ldl 0\n");
    fprintf(f,"?p %s ldl 1\n",flag);
    fprintf(f,"loop:\n");
    fprintf(f,"      jmri loop\n");
    
    fclose(f);
    
    char buffer[1024];
    sprintf(buffer,"./asm/asm < %s > %s 2> /dev/null",asm_filename,obj_filename);
    int res=system(buffer);
    assert(res==0);
    
    sprintf(buffer,"./simulator/simu -nox -ql -r0 %s > /dev/null;echo $? > %s",obj_filename,out_filename);
    res=system(buffer);
    
    f=fopen(out_filename,"r");
    if(fscanf(f,"%d",&res)!=1)
    {
        fclose(f);
        return false;
    }
    fclose(f);
    
    if(res!=0 && res!=1)
        return false;
    
    bool real_result=(res==1);
    
    return real_result==expected_result;
}

int decode(int a)
{
    uint16_t b=static_cast<uint16_t>(a);
    int16_t c=static_cast<int16_t>(b);
    return c;
}

bool le(int a,int b)
{
    return decode(a)<=decode(b);
}

bool lt(int a,int b)
{
    return decode(a)<decode(b);
}

bool ge(int a,int b)
{
    return decode(a)>=decode(b);
}

bool gt(int a,int b)
{
    return decode(a)>decode(b);
}

bool eq(int a,int b)
{
    return decode(a)==decode(b);
}

bool ne(int a,int b)
{
    return decode(a)!=decode(b);
}

void test_cmp()
{
    printf("* Testing compare...\n");
    printf("* * Testing LE...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"cmp","LE",le(i,j));
        if(!res)
            printf("error: cmp/LE/%#x,%#x failed\n",i,j);
    }
    printf("* * Testing LT...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"cmp","LT",lt(i,j));
        if(!res)
            printf("error: cmp/LT/%#x,%#x failed\n",i,j);
    }
    printf("* * Testing GE...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"cmp","GE",ge(i,j));
        if(!res)
            printf("error: cmp/GE/%#x,%#x failed\n",i,j);
    }
    printf("* * Testing GT...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"cmp","GT",gt(i,j));
        if(!res)
            printf("error: cmp/GT/%#x,%#x failed\n",i,j);
    }
    printf("* * Testing EQ...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"cmp","EQ",eq(i,j));
        if(!res)
            printf("error: cmp/EQ/%#x,%#x failed\n",i,j);
    }
    printf("* * Testing NE...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"cmp","NE",ne(i,j));
        if(!res)
            printf("error: cmp/NE/%#x,%#x failed\n",i,j);
    }
}

void test_sub()
{
    printf("* Testing subtract...\n");
    printf("* * Testing LE...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"sub","LE",le(i,j));
        if(!res)
            printf("error: cmp/LE/%#x,%#x failed\n",i,j);
    }
    printf("* * Testing LT...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"sub","LT",lt(i,j));
        if(!res)
            printf("error: cmp/LT/%#x,%#x failed\n",i,j);
    }
    printf("* * Testing GE...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"sub","GE",ge(i,j));
        if(!res)
            printf("error: cmp/GE/%#x,%#x failed\n",i,j);
    }
    printf("* * Testing GT...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"sub","GT",gt(i,j));
        if(!res)
            printf("error: cmp/GT/%#x,%#x failed\n",i,j);
    }
    printf("* * Testing EQ...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"sub","EQ",eq(i,j));
        if(!res)
            printf("error: cmp/EQ/%#x,%#x failed\n",i,j);
    }
    printf("* * Testing NE...\n");
    for(int k=0;k<100;k++)
    {
        int i=rand()%(0xffff+1);
        int j=rand()%(0xffff+1);
        bool res=execute_and_test(i,j,"sub","NE",ne(i,j));
        if(!res)
            printf("error: cmp/NE/%#x,%#x failed\n",i,j);
    }
}

int main()
{
    srand(time(NULL));
    test_cmp();
    test_sub();
    
    return 0;
}
